# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("show", "parse_display_args")

import sys
from collections.abc import Sequence
from shutil import get_terminal_size
from typing import Any, Literal, TextIO

import networkx

from .._nodes import NodeKey
from .._pipeline_graph import PipelineGraph
from .._tasks import TaskInitNode, TaskNode
from ._formatting import GetNodeText, get_node_symbol
from ._layout import ColumnSelector, Layout
from ._merge import (
    MergedNodeKey,
    merge_graph_input_trees,
    merge_graph_intermediates,
    merge_graph_output_trees,
)
from ._options import NodeAttributeOptions
from ._printer import make_default_printer

DisplayNodeKey = NodeKey | MergedNodeKey


def parse_display_args(
    pipeline_graph: PipelineGraph,
    *,
    dataset_types: bool = False,
    init: bool | None = False,
    dimensions: Literal["full"] | Literal["concise"] | Literal[False] | None = None,
    task_classes: Literal["full"] | Literal["concise"] | Literal[False] = False,
    storage_classes: bool | None = None,
    merge_input_trees: int = 4,
    merge_output_trees: int = 4,
    merge_intermediates: bool = True,
    include_automatic_connections: bool = False,
) -> tuple[networkx.DiGraph | networkx.MultiDiGraph, NodeAttributeOptions]:
    """Print a text-based ~.PipelineGraph` visualization.

    Parameters
    ----------
    pipeline_graph : `PipelineGraph`
        Graph to display.
    dataset_types : `bool`, optional
        Whether to include dataset type nodes (default is `False`).
    init : `bool`, optional
        Whether to include task initialization nodes (i.e. the producers of
        init-input and init-output dataset types).  Default is `False`.  `None`
        will show both runtime and init nodes, while `True` will show only
        init nodes.
    dimensions : `str` or `False`, optional
        How to display the dimensions of tasks and dataset types.

        Possible values include:

        - ``"full"``: report fully-expanded dimensions.
        - ``"concise"``: report only dimensions that are not required or
          implied dependencies of any reported dimension.
        - `False`: do not report dimensions at all.
        - `None` (default): report concise dimensions only if
          `PipelineGraph.is_fully_resolved` is `True`.

        This also sets whether merge options consider dimensions when merging
        nodes.
    task_classes : `str` or `False`, optional
        How to display task class names (not task labels, which are always
        shown).

        Possible values include:

        - ``"full"``: report the fully-qualified task class name.
        - ``"concise"``: report the task class name with no module
          or package.
        - `False`: (default) do not report task classes at all.

        This also sets whether merge options consider task classes when merging
        nodes.  It is ignored if ``tasks=False``.
    storage_classes : `bool`, optional
        Whether to display storage classes in dataset type nodes. This also
        sets whether merge options consider storage classes when merging nodes.
        It is ignored if ``dataset_types=False``.
    merge_input_trees : `int`, optional
        If positive, merge input trees of the graph whose nodes have the same
        outputs and other properties (dimensions, task classes, storage
        classes), traversing this many nodes deep into the graph from the
        beginning.  Default is ``4``.
    merge_output_trees : `int`, optional
        If positive, merge output trees of the graph whose nodes have the same
        outputs and other properties (dimensions, task classes, storage
        classes), traversing this many nodes deep into the graph from the end.
        Default is ``4``.
    merge_intermediates : `bool`, optional
        If `True` (default) merge interior parallel nodes with the same inputs,
        outputs, and other properties (dimensions, task classes, storage
        classes).
    include_automatic_connections : `bool`, optional
        Whether to include automatically-added connections like the config,
        log, and metadata dataset types for each task.  Default is `False`.
    """
    if init is None:
        if not dataset_types:
            raise ValueError("Cannot show init and runtime graphs unless dataset types are shown.")
        xgraph = pipeline_graph.make_xgraph()
    elif dataset_types:
        xgraph = pipeline_graph.make_bipartite_xgraph(init)
    else:
        xgraph = pipeline_graph.make_task_xgraph(init)
        storage_classes = False

    options = NodeAttributeOptions(
        dimensions=dimensions, storage_classes=storage_classes, task_classes=task_classes
    )
    options = options.checked(pipeline_graph.is_fully_resolved)

    if dataset_types and not include_automatic_connections:
        taskish_nodes: list[TaskNode | TaskInitNode] = []
        for task_node in pipeline_graph.tasks.values():
            if init is None or init is False:
                taskish_nodes.append(task_node)
            if init is None or init is True:
                taskish_nodes.append(task_node.init)
        for t in taskish_nodes:
            xgraph.remove_nodes_from(
                edge.dataset_type_key
                for edge in t.iter_all_outputs()
                if edge.connection_name not in t.outputs and not xgraph.out_degree(edge.dataset_type_key)
            )

    if merge_input_trees:
        merge_graph_input_trees(xgraph, options, depth=merge_input_trees)
    if merge_output_trees:
        merge_graph_output_trees(xgraph, options, depth=merge_output_trees)
    if merge_intermediates:
        merge_graph_intermediates(xgraph, options)

    return xgraph, options


def show(
    pipeline_graph: PipelineGraph,
    stream: TextIO = sys.stdout,
    *,
    color: bool | Sequence[str] | None = None,
    width: int = -1,
    column_interior_penalty: int = 1,
    column_crossing_penalty: int = 1,
    column_insertion_penalty: int = 2,
    **kwargs: Any,
) -> None:
    """Print a text-based ~.PipelineGraph` visualization.

    Parameters
    ----------
    pipeline_graph : `PipelineGraph`
        Graph to display.
    stream : `TextIO`, optional
        Output stream.  Defaults to STDOUT.
    color : `bool` or `~collections.abc.Sequence` [ `str` ]
        Whether to use tto add color to the graph. Default is to add colors
        only if the `colorama` package can be imported. `False` disables colors
        unconditionally, while `True` or a sequence of colors (see
        `make_colorama_printer`) will result in `ImportError` being propagated
        up if `colorama` is unavailable.
    width : `int`, optional
        Number of columns the full graph should occupy, including text
        descriptions on the right.  If ``0``, there is no limit (and hence no
        text-wrapping).  If negative (default) use the current terminal width.
    column_interior_penalty : `int`
        Penalty applied to a prospective column for a node when that column is
        between two existing columns.
    column_crossing_penalty : `int`
        Penalty applied to a prospective column for a node for each ongoing
        (vertical) edge that node's incoming edges would have to "hop".
    column_insertion_penalty : `int`
        Penalty applied to a prospective column for a node when considering a
        new columns on the sides or between two existing columns.
    **kwargs
        Forwarded to parse_display_args.
    """
    xgraph, options = parse_display_args(pipeline_graph, **kwargs)

    column_selector = ColumnSelector(
        interior_penalty=column_interior_penalty,
        crossing_penalty=column_crossing_penalty,
        insertion_penalty=column_insertion_penalty,
    )
    layout = Layout[DisplayNodeKey](xgraph, column_selector)

    if width < 0:
        width, _ = get_terminal_size()

    printer = make_default_printer(layout.width, color, stream)
    printer.get_symbol = get_node_symbol

    get_text = GetNodeText(xgraph, options, (width - printer.width) if width else 0)
    printer.get_text = get_text

    printer.print(stream, layout)
    for line in get_text.format_deferrals(width):
        print(line, file=stream)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("get_node_symbol", "GetNodeText", "format_dimensions", "format_task_class")

import itertools
import textwrap
from collections.abc import Iterator

import networkx
import networkx.algorithms.community
from lsst.daf.butler import DimensionGroup

from .._nodes import NodeKey, NodeType
from ._merge import MergedNodeKey
from ._options import NodeAttributeOptions

DisplayNodeKey = NodeKey | MergedNodeKey
"""Type alias for graph keys that may be original task, task init, or dataset
type keys, or a merge of several keys for display purposes.
"""


def get_node_symbol(node: DisplayNodeKey, x: int | None = None) -> str:
    """Return a single-character symbol for a particular node type.

    Parameters
    ----------
    node : `DisplayNodeKey`
        Named tuple used as the node key.
    x : `str`, optional
        Ignored; may be passed for compatibility with the `Printer` class's
        ``get_symbol`` callback.

    Returns
    -------
    symbol : `str`
        A single-character symbol.
    """
    match node:
        case NodeKey(node_type=NodeType.TASK):
            return "■"
        case NodeKey(node_type=NodeType.DATASET_TYPE):
            return "○"
        case NodeKey(node_type=NodeType.TASK_INIT):
            return "▣"
        case MergedNodeKey(node_type=NodeType.TASK):
            return "▥"
        case MergedNodeKey(node_type=NodeType.DATASET_TYPE):
            return "◍"
        case MergedNodeKey(node_type=NodeType.TASK_INIT):
            return "▤"
    raise ValueError(f"Unexpected node key: {node} of type {type(node)}.")


class GetNodeText:
    """A callback for the `Printer` class's `get_text` callback that
    prints detailed information about a node and can defer long entries to
    a footnote.

    Parameters
    ----------
    xgraph : `networkx.DiGraph` or `networkx.MultiDiGraph`
        NetworkX export of a `.PipelineGraph` that is being displayed.
    options : `NodeAttributeOptions`
        Options for how much information to display.
    width : `int` or `None`
        Number of display columns that node text can occupy.  `None` for
        unlimited.
    """

    def __init__(
        self,
        xgraph: networkx.DiGraph | networkx.MultiDiGraph,
        options: NodeAttributeOptions,
        width: int | None,
    ):
        self.xgraph = xgraph
        self.options = options
        self.width = width
        self.deferred: list[tuple[str, tuple[str, str], list[str]]] = []

    def __call__(self, node: DisplayNodeKey, x: int, style: tuple[str, str]) -> str:
        state = self.xgraph.nodes[node]
        terms: list[str] = [f"{node}:" if self.options else str(node)]
        if self.options.dimensions and node.node_type != NodeType.TASK_INIT:
            terms.append(self.format_dimensions(state["dimensions"]))
        if self.options.task_classes and (
            node.node_type is NodeType.TASK or node.node_type is NodeType.TASK_INIT
        ):
            terms.append(self.format_task_class(state["task_class_name"]))
        if self.options.storage_classes and node.node_type is NodeType.DATASET_TYPE:
            terms.append(state["storage_class_name"])
        description = " ".join(terms)
        if self.width and len(description) > self.width:
            index = f"[{len(self.deferred) + 1}]"
            self.deferred.append((index, style, terms))
            return f"{description[:self.width - len(index) - 6]}...{style[0]}{index}{style[1]} "
        return description

    def format_dimensions(self, dimensions: DimensionGroup) -> str:
        """Format the dimensions of a task or dataset type node.

        Parameters
        ----------
        dimensions : `~lsst.daf.butler.DimensionGroup`
            The dimensions to be formatted.

        Returns
        -------
        formatted : `str`
            The formatted dimension string.
        """
        return format_dimensions(self.options, dimensions)

    def format_task_class(self, task_class_name: str) -> str:
        """Format the type object for a task or task init node.

        Parameters
        ----------
        task_class_name : `str`
            The name of the task class.

        Returns
        -------
        formatted : `str`
            The formatted string.
        """
        return format_task_class(self.options, task_class_name)

    def format_deferrals(self, width: int | None) -> Iterator[str]:
        """Iterate over all descriptions that were truncated earlier and
        replace with footnote placeholders.

        Parameters
        ----------
        width : `int` or `None`
            Number of columns to wrap descriptions at.

        Returns
        -------
        deferrals : `collections.abc.Iterator` [ `str` ]
            Lines of deferred text, already wrapped.
        """
        indent = "  "
        for index, style, terms in self.deferred:
            yield f"{style[0]}{index}{style[1]}"
            for term in terms:
                if width:
                    yield from textwrap.wrap(term, width, initial_indent=indent, subsequent_indent=indent)
                else:
                    yield term


def format_dimensions(options: NodeAttributeOptions, dimensions: DimensionGroup) -> str:
    """Format the dimensions of a task or dataset type node.

    Parameters
    ----------
    options : `NodeAttributeOptions`
        Options for how much information to display.
    dimensions : `~lsst.daf.butler.DimensionGroup`
        The dimensions to be formatted.

    Returns
    -------
    formatted : `str`
        The formatted dimension string.
    """
    match options.dimensions:
        case "full":
            return str(dimensions.names)
        case "concise":
            redundant: set[str] = set()
            for a, b in itertools.permutations(dimensions.required, 2):
                if a in dimensions.universe[b].required:
                    redundant.add(a)
            kept = [d for d in dimensions.required if d not in redundant]
            assert dimensions.universe.conform(kept) == dimensions
            return f"{{{', '.join(kept)}}}"
        case False:
            return ""
    raise ValueError(f"Invalid display option for dimensions: {options.dimensions!r}.")


def format_task_class(options: NodeAttributeOptions, task_class_name: str) -> str:
    """Format the type object for a task or task init node.

    Parameters
    ----------
    options : `NodeAttributeOptions`
        Options for how much information to display.
    task_class_name : `str`
        The name of the task class.

    Returns
    -------
    formatted : `str`
        The formatted string.
    """
    match options.task_classes:
        case "full":
            return task_class_name
        case "concise":
            return task_class_name.split(".")[-1]
        case False:
            return ""
    raise ValueError(f"Invalid display option for task_classes: {options.task_classes!r}.")



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for Pipeline.
"""

import pickle
import textwrap
import unittest

import lsst.utils.tests
from lsst.pipe.base import LabelSpecifier, Pipeline, TaskDef
from lsst.pipe.base.pipelineIR import LabeledSubset
from lsst.pipe.base.tests.simpleQGraph import AddTask, makeSimplePipeline


class PipelineTestCase(unittest.TestCase):
    """A test case for TaskDef and Pipeline."""

    def testTaskDef(self):
        """Tests for TaskDef structure."""
        task1 = TaskDef(taskClass=AddTask, config=AddTask.ConfigClass())
        self.assertIn("Add", task1.taskName)
        self.assertIsInstance(task1.config, AddTask.ConfigClass)
        self.assertIsNotNone(task1.taskClass)
        self.assertEqual(task1.label, "add_task")
        task1a = pickle.loads(pickle.dumps(task1))
        self.assertEqual(task1, task1a)

    def testEmpty(self):
        """Creating empty pipeline."""
        pipeline = Pipeline("test")
        self.assertEqual(len(pipeline), 0)

    def testInitial(self):
        """Testing constructor with initial data."""
        pipeline = makeSimplePipeline(2)
        self.assertEqual(len(pipeline), 2)
        pipeline_graph = pipeline.to_graph()
        pipeline_graph.sort()
        task_nodes = list(pipeline_graph.tasks.values())
        self.assertEqual(task_nodes[0].task_class, AddTask)
        self.assertEqual(task_nodes[1].task_class, AddTask)
        self.assertEqual(task_nodes[0].label, "task0")
        self.assertEqual(task_nodes[1].label, "task1")
        self.assertEqual(pipeline.task_labels, {"task0", "task1"})

    def testModifySubset(self):
        pipeline = makeSimplePipeline(2)

        # Test adding labels.
        with self.assertRaises(ValueError):
            pipeline.addLabelToSubset("test", "new_label")
        pipeline._pipelineIR.labeled_subsets["test"] = LabeledSubset("test", set(), None)
        with self.assertRaises(ValueError):
            pipeline.addLabelToSubset("test", "missing_label")
        pipeline.addLabelToSubset("test", "task0")
        self.assertEqual(pipeline._pipelineIR.labeled_subsets["test"].subset, {"task0"})

        # Test removing labels.
        with self.assertRaises(ValueError):
            pipeline.addLabelToSubset("missing_subset", "task0")
        with self.assertRaises(ValueError):
            pipeline.addLabelToSubset("test", "missing_label")
        pipeline.removeLabelFromSubset("test", "task0")
        self.assertEqual(pipeline._pipelineIR.labeled_subsets["test"].subset, set())

        # Test creating new labeled subsets.
        with self.assertRaises(ValueError):
            # missing task label
            pipeline.addLabeledSubset("newSubset", "test description", {"missing_task_label"})
        with self.assertRaises(ValueError):
            # duplicate labeled subset
            pipeline.addLabeledSubset("test", "test description", {"missing_task_label"})

        taskLabels = {"task0", "task1"}
        pipeline.addLabeledSubset("newSubset", "test description", taskLabels)

        # verify using the subset property interface
        self.assertEqual(pipeline.subsets["newSubset"], taskLabels)

        # Test removing labeled subsets
        with self.assertRaises(ValueError):
            pipeline.removeLabeledSubset("missing_subset")

        pipeline.removeLabeledSubset("newSubset")
        self.assertNotIn("newSubset", pipeline.subsets.keys())

        pipeline.addLabeledSubset("testSubset", "Test subset description", taskLabels)
        taskSubset = {"task0"}
        pipelineDrop = pipeline.subsetFromLabels(LabelSpecifier(taskSubset), pipeline.PipelineSubsetCtrl.DROP)
        pipelineEdit = pipeline.subsetFromLabels(LabelSpecifier(taskSubset), pipeline.PipelineSubsetCtrl.EDIT)

        # Test subsetting from labels
        self.assertNotIn(taskLabels - taskSubset, set(pipelineDrop.task_labels))
        self.assertNotIn("testSubset", pipelineDrop.subsets.keys())
        self.assertNotIn(taskLabels - taskSubset, set(pipelineEdit.task_labels))
        self.assertIn("testSubset", pipelineEdit.subsets.keys())
        self.assertEqual(pipelineEdit.subsets["testSubset"], taskSubset)

    def testMergingPipelines(self):
        pipeline1 = makeSimplePipeline(2)
        pipeline2 = makeSimplePipeline(4)
        pipeline2.removeTask("task0")
        pipeline2.removeTask("task1")

        pipeline1.mergePipeline(pipeline2)
        self.assertEqual(pipeline1._pipelineIR.tasks.keys(), {"task0", "task1", "task2", "task3"})

    def testFindingSubset(self):
        pipeline = makeSimplePipeline(2)
        pipeline._pipelineIR.labeled_subsets["test1"] = LabeledSubset("test1", set(), None)
        pipeline._pipelineIR.labeled_subsets["test2"] = LabeledSubset("test2", set(), None)
        pipeline._pipelineIR.labeled_subsets["test3"] = LabeledSubset("test3", set(), None)

        pipeline.addLabelToSubset("test1", "task0")
        pipeline.addLabelToSubset("test3", "task0")

        with self.assertRaises(ValueError):
            pipeline.findSubsetsWithLabel("missing_label")

        self.assertEqual(pipeline.findSubsetsWithLabel("task0"), {"test1", "test3"})

    def testParameters(self):
        """Test that parameters can be set and used to format."""
        pipeline_str = textwrap.dedent(
            """
            description: Test Pipeline
            parameters:
               testValue: 5
            tasks:
              add:
                class: test_pipeline.AddTask
                config:
                  addend: parameters.testValue
        """
        )
        # verify that parameters are used in expanding a pipeline
        pipeline = Pipeline.fromString(pipeline_str)
        pipeline_graph = pipeline.to_graph()
        self.assertEqual(pipeline_graph.tasks["add"].config.addend, 5)

        # verify that a parameter can be overridden on the "command line"
        pipeline.addConfigOverride("parameters", "testValue", 14)
        pipeline_graph = pipeline.to_graph()
        self.assertEqual(pipeline_graph.tasks["add"].config.addend, 14)

        # verify that parameters does not support files or python overrides
        with self.assertRaises(ValueError):
            pipeline.addConfigFile("parameters", "fakeFile")
        with self.assertRaises(ValueError):
            pipeline.addConfigPython("parameters", "fakePythonString")

    def testSerialization(self):
        pipeline = makeSimplePipeline(2)
        dump = str(pipeline)
        load = Pipeline.fromString(dump)
        self.assertEqual(pipeline, load)


class MyMemoryTestCase(lsst.utils.tests.MemoryTestCase):
    """Run file leak tests."""


def setup_module(module):
    """Configure pytest."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



from . import automatic_connection_constants, connectionTypes, pipeline_graph, pipelineIR, utils
from ._dataset_handle import *

# Symbols from _datasetQueryConstraints are exported from
# all_dimensions_quantum_graph_builder, since that's the only place they are
# used.
from ._instrument import *
from ._observation_dimension_packer import *
from ._quantumContext import *
from ._status import *
from ._task_metadata import *
from .config import *
from .connections import *
from .executionButlerBuilder import *
from .graph import *
from .pipeline import *

# We import the main PipelineGraph type and the module (above), but we don't
# lift all symbols to package scope.
from .pipeline_graph import PipelineGraph
from .pipelineTask import *
from .struct import *
from .task import *
from .taskFactory import *
from .version import *

# quantum_graph_builder, all_dimensions_quantum_graph_builder,
# quantum_graph_skeleton, and prerequisite_helper symbols are intentionally not
# lifted to package scope.



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining config classes for PipelineTask.
"""

from __future__ import annotations

__all__ = ["PipelineTaskConfig"]

# -------------------------------
#  Imports of standard modules --
# -------------------------------
import os
from collections.abc import Iterable
from numbers import Number
from typing import TYPE_CHECKING, Any, TypeVar

# -----------------------------
#  Imports for other modules --
# -----------------------------
import lsst.pex.config as pexConfig

from ._instrument import Instrument
from .configOverrides import ConfigOverrides
from .connections import PipelineTaskConnections
from .pipelineIR import ConfigIR, ParametersIR

if TYPE_CHECKING:
    from lsst.pex.config.callStack import StackFrame

# ----------------------------------
#  Local non-exported definitions --
# ----------------------------------

_S = TypeVar("_S", bound="PipelineTaskConfigMeta")

# ------------------------
#  Exported definitions --
# ------------------------


class TemplateField(pexConfig.Field):
    """Field specialized for use with connection templates.

    Specifically it treats strings or numbers as valid input, as occasionally
    numbers are used as a cycle counter in templates.

    The reason for the specialized field, is that when numbers are involved
    with the config override system through pipelines or from the command line,
    sometimes the quoting to get appropriate values as strings gets
    complicated. This will simplify the process greatly.
    """

    def _validateValue(self, value: Any) -> None:
        if value is None:
            return

        if not (isinstance(value, str | Number)):
            raise TypeError(
                f"Value {value} is of incorrect type {pexConfig.config._typeStr(value)}."
                " Expected type str or a number"
            )
        if self.check is not None and not self.check(value):
            ValueError("Value {value} is not a valid value")

    def __set__(
        self,
        instance: pexConfig.Config,
        value: Any,
        at: StackFrame | None = None,
        label: str = "assignment",
    ) -> None:
        # validate first, even though validate will be called in super
        self._validateValue(value)
        # now, explicitly make it into a string
        value = str(value)
        super().__set__(instance, value, at, label)


class PipelineTaskConfigMeta(pexConfig.ConfigMeta):
    """Metaclass used in the creation of PipelineTaskConfig classes.

    This metaclass ensures a `PipelineTaskConnections` class is specified in
    the class construction parameters with a parameter name of
    pipelineConnections. Using the supplied connection class, this metaclass
    constructs a `lsst.pex.config.Config` instance which can be used to
    configure the connections class. This config is added to the config class
    under declaration with the name "connections" used as an identifier. The
    connections config also has a reference to the connections class used in
    its construction associated with an atttribute named ``ConnectionsClass``.
    Finally the newly constructed config class (not an instance of it) is
    assigned to the Config class under construction with the attribute name
    ``ConnectionsConfigClass``.

    Parameters
    ----------
    name : `str`
        Name of config.
    bases : `~collections.abc.Collection`
        Base classes.
    dct : `~collections.abc.Mapping`
        Parameter dict.
    **kwargs : `~typing.Any`
        Additional parameters.
    """

    def __new__(
        cls: type[_S],
        name: str,
        bases: tuple[type[PipelineTaskConfig], ...],
        dct: dict[str, Any],
        **kwargs: Any,
    ) -> _S:
        if name != "PipelineTaskConfig":
            # Verify that a connection class was specified and the argument is
            # an instance of PipelineTaskConfig
            if "pipelineConnections" not in kwargs:
                for base in bases:
                    if hasattr(base, "connections"):
                        kwargs["pipelineConnections"] = base.connections.dtype.ConnectionsClass
                        break
            if "pipelineConnections" not in kwargs:
                raise NameError("PipelineTaskConfig or a base class must be defined with connections class")
            connectionsClass = kwargs["pipelineConnections"]
            if not issubclass(connectionsClass, PipelineTaskConnections):
                raise ValueError("Can only assign a PipelineTaskConnections Class to pipelineConnections")

            # Create all the fields that will be used in the newly created sub
            # config (under the attribute name "connections")
            configConnectionsNamespace: dict[str, pexConfig.Field] = {}
            for fieldName, obj in connectionsClass.allConnections.items():
                configConnectionsNamespace[fieldName] = pexConfig.Field[str](
                    doc=f"name for connection {fieldName}", default=obj.name, deprecated=obj.deprecated
                )
            # If there are default templates also add them as fields to
            # configure the template values
            if hasattr(connectionsClass, "defaultTemplates"):
                docString = "Template parameter used to format corresponding field template parameter"
                for templateName, default in connectionsClass.defaultTemplates.items():
                    configConnectionsNamespace[templateName] = TemplateField(
                        dtype=str,
                        doc=docString,
                        default=default,
                        deprecated=connectionsClass.deprecatedTemplates.get(templateName),
                    )
            # add a reference to the connection class used to create this sub
            # config
            configConnectionsNamespace["ConnectionsClass"] = connectionsClass

            # Create a new config class with the fields defined above
            Connections = type(f"{name}Connections", (pexConfig.Config,), configConnectionsNamespace)
            # add it to the Config class that is currently being declared
            dct["connections"] = pexConfig.ConfigField(
                dtype=Connections,
                doc="Configurations describing the connections of the PipelineTask to datatypes",
            )
            dct["ConnectionsConfigClass"] = Connections
            dct["ConnectionsClass"] = connectionsClass
        inst = super().__new__(cls, name, bases, dct)
        return inst

    def __init__(
        self, name: str, bases: tuple[type[PipelineTaskConfig], ...], dct: dict[str, Any], **kwargs: Any
    ):
        # This overrides the default init to drop the kwargs argument. Python
        # metaclasses will have this argument set if any kwargs are passes at
        # class construction time, but should be consumed before calling
        # __init__ on the type metaclass. This is in accordance with python
        # documentation on metaclasses
        super().__init__(name, bases, dct)

    ConnectionsClass: type[PipelineTaskConnections]
    ConnectionsConfigClass: type[pexConfig.Config]


class PipelineTaskConfig(pexConfig.Config, metaclass=PipelineTaskConfigMeta):
    """Configuration class for `PipelineTask`.

    This Configuration class functions in largely the same manner as any other
    derived from `lsst.pex.config.Config`. The only difference is in how it is
    declared. `PipelineTaskConfig` children need to be declared with a
    pipelineConnections argument. This argument should specify a child class of
    `PipelineTaskConnections`. During the declaration of a `PipelineTaskConfig`
    a config class is created with information from the supplied connections
    class to allow configuration of the connections class. This dynamically
    created config class is then attached to the `PipelineTaskConfig` via a
    `~lsst.pex.config.ConfigField` with the attribute name ``connections``.
    """

    connections: pexConfig.ConfigField
    """Field which refers to a dynamically added configuration class which is
    based on a PipelineTaskConnections class.
    """

    saveLogOutput = pexConfig.Field[bool](
        default=True,
        optional=False,
        doc="Flag to enable/disable saving of log output for a task, enabled by default.",
    )

    def applyConfigOverrides(
        self,
        instrument: Instrument | None,
        taskDefaultName: str,
        pipelineConfigs: Iterable[ConfigIR] | None,
        parameters: ParametersIR,
        label: str,
    ) -> None:
        """Apply config overrides to this config instance.

        Parameters
        ----------
        instrument : `Instrument` or `None`
            An instance of the `Instrument` specified in a pipeline.
            If `None` then the pipeline did not specify and instrument.
        taskDefaultName : `str`
            The default name associated with the `Task` class. This
            may be used with instrumental overrides.
        pipelineConfigs : `~collections.abc.Iterable` \
                of `~.pipelineIR.ConfigIR`
            An iterable of `~.pipelineIR.ConfigIR` objects that contain
            overrides to apply to this config instance.
        parameters : `~.pipelineIR.ParametersIR`
            Parameters defined in a Pipeline which are used in formatting
            of config values across multiple `Task` in a pipeline.
        label : `str`
            The label associated with this class's Task in a pipeline.
        """
        overrides = ConfigOverrides()
        overrides.addParameters(parameters)
        if instrument is not None:
            overrides.addInstrumentOverride(instrument, taskDefaultName)
        if pipelineConfigs is not None:
            for subConfig in (configIr.formatted(parameters) for configIr in pipelineConfigs):
                if subConfig.dataId is not None:
                    raise NotImplementedError(
                        "Specializing a config on a partial data id is not yet "
                        "supported in Pipeline definition"
                    )
                # only apply override if it applies to everything
                if subConfig.dataId is None:
                    if subConfig.file:
                        for configFile in subConfig.file:
                            overrides.addFileOverride(os.path.expandvars(configFile))
                    if subConfig.python is not None:
                        overrides.addPythonOverride(subConfig.python)
                    for key, value in subConfig.rest.items():
                        overrides.addValueOverride(key, value)
        overrides.applyTo(self)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("show_dot",)

import html
import os
import sys
from collections.abc import Mapping
from typing import Any, TextIO

from .._nodes import NodeType
from .._pipeline_graph import PipelineGraph
from ._formatting import NodeKey, format_dimensions, format_task_class
from ._options import NodeAttributeOptions
from ._show import parse_display_args

_ATTRIBS = {
    NodeType.TASK: dict(style="filled", color="black", fillcolor="#B1F2EF"),
    NodeType.DATASET_TYPE: dict(style="rounded,filled,bold", color="#00BABC", fillcolor="#F5F5F5"),
    NodeType.TASK_INIT: dict(style="filled", color="black", fillcolor="#F4DEFA"),
}
_DEFAULT_GRAPH = dict(splines="ortho", nodesep="0.5", ranksep="0.75")
_DEFAULT_NODE = dict(shape="box", fontname="Monospace", fontsize="14", margin="0.2,0.1", penwidth="3")
_DEFAULT_EDGE = dict(color="black", arrowsize="1.5", penwidth="1.5", pad="10mm")
_LABEL_POINT_SIZE = "18"
_LABEL_MAX_LINES_SOFT = 10
_LABEL_MAX_LINES_HARD = 15
_OVERFLOW_MAX_LINES = 20


def show_dot(
    pipeline_graph: PipelineGraph,
    stream: TextIO = sys.stdout,
    **kwargs: Any,
) -> None:
    """Write a DOT representation of the pipeline graph to a stream.

    Parameters
    ----------
    pipeline_graph : `PipelineGraph`
        Pipeline graph to show.
    stream : `TextIO`, optional
        Stream to write the DOT representation to.
    **kwargs
        Additional keyword arguments to pass to `parse_display_args`.
    """
    xgraph, options = parse_display_args(pipeline_graph, **kwargs)

    print("digraph Pipeline {", file=stream)
    _render_default("graph", _DEFAULT_GRAPH, stream)
    _render_default("node", _DEFAULT_NODE, stream)
    _render_default("edge", _DEFAULT_EDGE, stream)

    overflow_ref = 1
    overflow_ids = []
    for node_key, node_data in xgraph.nodes.items():
        match node_key.node_type:
            case NodeType.TASK | NodeType.TASK_INIT:
                _render_task_node(node_key, node_data, options, stream)
            case NodeType.DATASET_TYPE:
                overflow_ref, node_overflow_ids = _render_dataset_type_node(
                    node_key, node_data, options, stream, overflow_ref
                )
                if node_overflow_ids:
                    overflow_ids += node_overflow_ids
            case _:
                raise AssertionError(f"Unexpected node type: {node_key.node_type}")

    if overflow_ids:
        formatted_overflow_ids = [f'"{overflow_id}"' for overflow_id in overflow_ids]
        print(f"{{rank=sink; {'; '.join(formatted_overflow_ids)};}}", file=stream)

    for from_node, to_node, *_ in xgraph.edges:
        if xgraph.nodes[from_node].get("is_prerequisite", False):
            edge_data = dict(style="dashed")
        else:
            edge_data = {}
        _render_edge(from_node.node_id, to_node.node_id, stream, **edge_data)

    print("}", file=stream)


def _render_default(type: str, attribs: dict[str, str], stream: TextIO) -> None:
    """Set default attributes for a given type."""
    default_attribs = ", ".join([f'{key}="{val}"' for key, val in attribs.items()])
    print(f"{type} [{default_attribs}];", file=stream)


def _render_task_node(
    node_key: NodeKey,
    node_data: Mapping[str, Any],
    options: NodeAttributeOptions,
    stream: TextIO,
) -> None:
    """Render a Graphviz node for a task.

    Parameters
    ----------
    node_key : NodeKey
        The key for the node
    node_data : Mapping[str, Any]
        The data associated with the node
    options : NodeAttributeOptions
        Options for rendering the node
    stream : TextIO
        The stream to write the node to
    """
    labels, *_ = _format_label(str(node_key))

    # Add the fully resolved task class name
    if options.task_classes and (node_key.node_type in (NodeType.TASK, NodeType.TASK_INIT)):
        labels.append(html.escape(format_task_class(options, node_data["task_class_name"])))

    # Append dimensions to the node
    if options.dimensions and node_key.node_type != NodeType.TASK_INIT:
        labels.append(
            f"<I>dimensions:</I>&nbsp;{html.escape(format_dimensions(options, node_data['dimensions']))}"
        )

    _render_node(node_key.node_id, node_key.node_type, labels, stream)


def _render_dataset_type_node(
    node_key: NodeKey,
    node_data: Mapping[str, Any],
    options: NodeAttributeOptions,
    stream: TextIO,
    overflow_ref: int = 1,
) -> tuple[int, list[str]]:
    """Render a Graphviz node for a dataset type.

    Parameters
    ----------
    node_key : NodeKey
        The key for the node
    node_data : Mapping[str, Any]
        The data associated with the node
    options : NodeAttributeOptions
        Options for rendering the node
    stream : TextIO
        The stream to write the node to

    Returns
    -------
    overflow_ref : int
        The reference number for the next overflow node
    overflow_ids : str | None
        The ID of the overflow node, if any
    """
    labels, label_extras, common_prefix = _format_label(str(node_key), _LABEL_MAX_LINES_SOFT)
    if len(labels) + len(label_extras) <= _LABEL_MAX_LINES_HARD:
        labels += label_extras
        label_extras = []
    if common_prefix:
        labels.insert(0, common_prefix)

    # Add a reference to a free-floating overflow node
    label_extras_grouped = {}
    if label_extras:
        overflow_to_text = f"and {len(label_extras)} more, continued in [{overflow_ref}]"
        labels.append(f'<B><FONT POINT-SIZE="{_LABEL_POINT_SIZE}">{overflow_to_text}</FONT></B>')
        for i in range(0, len(label_extras), _OVERFLOW_MAX_LINES):
            overflow_id = f"{node_key.node_id}_{overflow_ref}_{i}"
            overflow_label_extras = label_extras[i : i + _OVERFLOW_MAX_LINES]
            if common_prefix:
                overflow_label_extras.insert(0, common_prefix)
            overflow_label_extras.insert(
                0, f'<B><FONT POINT-SIZE="{_LABEL_POINT_SIZE}">[{overflow_ref}]</FONT></B>'
            )
            label_extras_grouped[overflow_id] = overflow_label_extras
        overflow_ref += 1

    # Append dimensions to the node
    if options.dimensions:
        labels.append(
            "<I>dimensions:</I>&nbsp;" + html.escape(format_dimensions(options, node_data["dimensions"]))
        )

    # Append storage class to the node
    if options.storage_classes:
        labels.append("<I>storage class:</I>&nbsp;" + html.escape(node_data["storage_class_name"]))

    _render_node(node_key.node_id, node_key.node_type, labels, stream)

    # Render the overflow nodes and invisible edges, if any
    if label_extras_grouped:
        for overflow_id, overflow_labels in label_extras_grouped.items():
            _render_node(overflow_id, node_key.node_type, overflow_labels, stream)
            _render_edge(node_key.node_id, overflow_id, stream, **{"style": "invis"})

    overflow_ids = list(label_extras_grouped.keys())
    return overflow_ref, overflow_ids


def _render_node(
    node_id: str,
    node_type: NodeType,
    labels: list[str],
    stream: TextIO,
) -> None:
    """Render a Graphviz node.

    Parameters
    ----------
    node_id : str
        The unique name of the node
    node_type : NodeType
        The type of the node
    labels : list[str]
        The label elements to display on the node
    stream : TextIO
        The stream to write the node to
    """
    label = "".join([f'<TR><TD ALIGN="LEFT">{element}</TD></TR>' for element in labels])
    attrib_dict = dict(_ATTRIBS[node_type], label=label)
    pre = '<<TABLE BORDER="0" CELLPADDING="5">'
    post = "</TABLE>>"
    attrib = ", ".join(
        [
            f'{key}="{val}"' if key != "label" else f"{key}={pre}{val}{post}"
            for key, val in attrib_dict.items()
        ]
    )
    print(f'"{node_id}" [{attrib}];', file=stream)


def _render_edge(from_node_id: str, to_node_id: str, stream: TextIO, **kwargs: Any) -> None:
    """Render GV edge

    Parameters
    ----------
    from_node_id : str
        The unique ID of the node the edge is coming from
    to_node_id : str
        The unique ID of the node the edge is going to
    stream : TextIO
        The stream to write the edge to
    kwargs : Any
        Additional keyword arguments to pass to the edge
    """
    if kwargs:
        attrib = ", ".join([f'{key}="{val}"' for key, val in kwargs.items()])
        print(f'"{from_node_id}" -> "{to_node_id}" [{attrib}];', file=stream)
    else:
        print(f'"{from_node_id}" -> "{to_node_id}";', file=stream)


def _format_label(
    label: str,
    max_lines: int = 10,
    min_common_prefix_len: int = 1000,
) -> tuple[list[str], list[str], str]:
    """Add HTML-style formatting to label text.

    Parameters
    ----------
    label : str
        The label text to parse
    max_lines : int, optional
        The maximum number of lines to display
    min_common_prefix_len : int, optional
        The minimum length of a common prefix to consider

    Returns
    -------
    labels : list[str]
        Parsed and formatted label text elements
    label_extras : list[str]
        Parsed and formatted overflow text elements, if any
    common_prefix : str
        The common prefix of the label text, if any
    """
    parsed_labels, parsed_label_extras, common_prefix = _parse_label(label, max_lines, min_common_prefix_len)
    if common_prefix:
        common_prefix = f'<B><FONT POINT-SIZE="{_LABEL_POINT_SIZE}">{common_prefix}:</FONT></B>'

    labels = []
    label_extras = []
    indent = "&nbsp;&nbsp;" if common_prefix else ""
    for element in parsed_labels:
        labels.append(f'<B><FONT POINT-SIZE="{_LABEL_POINT_SIZE}">{indent}{element}</FONT></B>')
    for element in parsed_label_extras:
        label_extras.append(f'<B><FONT POINT-SIZE="{_LABEL_POINT_SIZE}">{indent}{element}</FONT></B>')

    return labels, label_extras, common_prefix


def _parse_label(
    label: str,
    max_lines: int,
    min_common_prefix_len: int,
) -> tuple[list[str], list[str], str]:
    """Parse label text into label elements.

    Parameters
    ----------
    label : str
        The label text to parse
    max_lines : int, optional
        The maximum number of lines to return (-1 if a common prefix present)
    min_common_prefix_len : int, optional
        The minimum length of a common prefix to consider

    Returns
    -------
    labels : list[str]
        Parsed label text elements
    label_extras : list[str]
        Overflow text elements, if any
    common_prefix : str
        The common prefix of the label text, if any
    """
    labels = label.split(", ")

    if len(labels) > 3 and len(common_prefix := os.path.commonprefix(labels)) > min_common_prefix_len:
        final_underscore_index = common_prefix.rfind("_")
        if final_underscore_index > 0:
            # Only use common prefixes that end in an underscore. This prevents
            # prefixes that may equal an entire element. For example, the label
            # "srcMatchFull, srcMatch" would return "srcMatch" as a common
            # prefix, and the labels list would contain an empty label.
            common_prefix = common_prefix[: final_underscore_index + 1]
            labels = [element[len(common_prefix) :] for element in labels]
        else:
            common_prefix = ""
    else:
        common_prefix = ""

    if (len(labels) + bool(common_prefix)) > max_lines:
        label_extras = labels[max_lines - bool(common_prefix) :]
        labels = labels[: max_lines - bool(common_prefix)]
    else:
        label_extras = []

    return labels, label_extras, common_prefix



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for execution_reports.
"""

import unittest

from lsst.pipe.base.execution_reports import QuantumGraphExecutionReport
from lsst.pipe.base.tests import simpleQGraph
from lsst.utils.tests import temporaryDirectory


class ExecutionReportsTestCase(unittest.TestCase):
    """Test execution reports.

    More tests are in lsst/ci_middleware/tests/test_prod_outputs.py
    """

    def test_make_reports(self) -> None:
        """Test make_reports to verify that output exists."""
        with temporaryDirectory() as root:
            # make a simple qgraph to make an execution report on
            butler, qgraph = simpleQGraph.makeSimpleQGraph(root=root)
            report = QuantumGraphExecutionReport.make_reports(butler, qgraph)
            # make a summary dictionary with a certain amount of
            # expected failures and check that they are there
            exp_failures = report.to_summary_dict(butler, do_store_logs=False)
            self.assertIsNotNone(exp_failures["task0"]["failed_quanta"])
            self.assertEqual(exp_failures["task1"]["outputs"]["add_dataset2"]["blocked"], 1)
            self.assertDictEqual(exp_failures["task2"]["failed_quanta"], {})
            self.assertEqual(exp_failures["task3"]["outputs"]["add_dataset4"]["produced"], 0)
            self.assertEqual(exp_failures["task4"]["n_quanta_blocked"], 1)
            self.assertIsNotNone(exp_failures["task0"]["n_expected"])
            # now we'll make a human-readable summary dict and
            # repeat the tests:
            hr_exp_failures = report.to_summary_dict(butler, do_store_logs=False, human_readable=True)
            self.assertIsNotNone(hr_exp_failures["task0"]["failed_quanta"])
            self.assertEqual(hr_exp_failures["task1"]["outputs"]["add_dataset2"]["blocked"], 1)
            self.assertListEqual(hr_exp_failures["task2"]["failed_quanta"], [])
            self.assertEqual(hr_exp_failures["task3"]["outputs"]["add_dataset4"]["produced"], 0)
            self.assertEqual(hr_exp_failures["task4"]["n_quanta_blocked"], 1)
            self.assertIsNotNone(hr_exp_failures["task0"]["n_expected"])



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["transfer_from_graph"]

from lsst.daf.butler import Butler, CollectionType, QuantumBackedButler, Registry
from lsst.daf.butler.registry import MissingCollectionError
from lsst.pipe.base import QuantumGraph


def transfer_from_graph(
    graph: str,
    dest: str,
    register_dataset_types: bool,
    transfer_dimensions: bool,
    update_output_chain: bool,
    dry_run: bool,
) -> int:
    """Transfer output datasets from quantum graph to dest.

    Parameters
    ----------
    graph : `str`
        URI string of the quantum graph.
    dest : `str`
        URI string of the destination Butler repo.
    register_dataset_types : `bool`
        Indicate whether missing dataset types should be registered.
    transfer_dimensions : `bool`
        Indicate whether dimensions should be transferred along with datasets.
        It can be more efficient to disable this if it is known that all
        dimensions exist.
    update_output_chain : `bool`
        If quantum graph metadata includes output run name and output
        collection which is a chain, update the chain definition to include run
        name as a the first collection in the chain.
    dry_run : `bool`
        Run the transfer without updating the destination butler.

    Returns
    -------
    count : `int`
        Actual count of transferred datasets.
    """
    # Read whole graph into memory
    qgraph = QuantumGraph.loadUri(graph)

    output_refs, _ = qgraph.get_refs(include_outputs=True, include_init_outputs=True, conform_outputs=True)

    # Get data repository dataset type definitions from the QuantumGraph.
    dataset_types = {dstype.name: dstype for dstype in qgraph.registryDatasetTypes()}

    # Make QBB, its config is the same as output Butler.
    qbb = QuantumBackedButler.from_predicted(
        config=dest,
        predicted_inputs=[ref.id for ref in output_refs],
        predicted_outputs=[],
        dimensions=qgraph.universe,
        datastore_records={},
        dataset_types=dataset_types,
    )

    dest_butler = Butler.from_config(dest, writeable=True)

    transferred = dest_butler.transfer_from(
        qbb,
        output_refs,
        transfer="auto",
        register_dataset_types=register_dataset_types,
        transfer_dimensions=transfer_dimensions,
        dry_run=dry_run,
    )
    count = len(transferred)

    # If anything was transferred then update output chain definition if asked.
    if count > 0 and update_output_chain and (metadata := qgraph.metadata) is not None:
        # These are defined in CmdLineFwk.
        output_run = metadata.get("output_run")
        output = metadata.get("output")
        input = metadata.get("input")
        if output_run is not None and output is not None:
            _update_chain(dest_butler.registry, output, output_run, input)

    return count


def _update_chain(registry: Registry, output_chain: str, output_run: str, inputs: list[str] | None) -> None:
    """Update chain definition if it exists to include run as the first item
    in a chain. If it does not exist then create it to include all inputs and
    output.
    """
    try:
        # If output_chain is not a chain the exception will be raised.
        chain_definition = list(registry.getCollectionChain(output_chain))
    except MissingCollectionError:
        # We have to create chained collection to include inputs and output run
        # (this reproduces logic in CmdLineFwk).
        registry.registerCollection(output_chain, type=CollectionType.CHAINED)
        chain_definition = list(registry.queryCollections(inputs, flattenChains=True)) if inputs else []
        chain_definition = [output_run] + [run for run in chain_definition if run != output_run]
        registry.setCollectionChain(output_chain, chain_definition)
    else:
        # If run is in the chain but not the first item then remove it, will
        # re-insert at front below.
        try:
            index = chain_definition.index(output_run)
            if index == 0:
                # It is already at the top.
                return
            else:
                del chain_definition[index]
        except ValueError:
            pass

        chain_definition.insert(0, output_run)
        registry.setCollectionChain(output_chain, chain_definition)
